/*
 *  Copyright (C) 2010, 2011, 2012 Michael Hofmann, Chemnitz University of Technology
 *  
 *  This file is part of the Sorting Library.
 *  
 *  The Sorting Library is free software: you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  The Sorting Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  

 *  
 *  SL - Sorting Library, michael <dot> hofmann <at> informatik <dot> tu-chemnitz <dot> de
 */


#include "sl_common.h"


slint_t elements_alloc_packed(packed_elements_t *s, slint_t nelements) /* sl_proto, sl_func elements_alloc_packed */
{
  if (s == NULL) return -1;

  s->size = s->max_size = 0;
  s->elements = NULL;

  if (nelements == 0) return 0;

  s->size = s->max_size = nelements;
  
  s->elements = z_alloc(nelements, pelem_byte);

  if (s->elements != NULL) return 0;

  elements_free_packed(s);

  return -1;
}


slint_t elements_free_packed(packed_elements_t *s) /* sl_proto, sl_func elements_free_packed */
{
  if (s == NULL) return -1;

  z_free(s->elements);

  s->size = s->max_size = 0;

  s->elements = NULL;

  return 0;
}


slint_t elements_alloc_packed_from_block(packed_elements_t *s, void *block, slint_t blocksize, slint_t alignment, slint_t nmax) /* sl_proto, sl_func elements_alloc_packed_from_block */
{
  if (s == NULL) return -1;

  s->size = s->max_size = blocksize / pelem_byte;

  if (nmax >= 0 && s->size > nmax) s->size = s->max_size = nmax;

  s->elements = block;

  return 0;
}


slint_t elements_pack_indexed(elements_t *s, packed_elements_t *d, slindex_t *rindx, slindex_t *windx) /* sl_proto, sl_func elements_pack_indexed */
{
  slint_t i;

  if (s->size > d->size) return -1;

#define DO_INDEXED \
  for (i = 0; i < s->size; ++i) \
  { \
    key_copy_at(s->keys, READ_INDEX(i), &d->elements[WRITE_INDEX(i)].key, 0); \
    data_copy_at(s, READ_INDEX(i), &d->elements[WRITE_INDEX(i)], 0); \
  }

  if (rindx == NULL)
  {
    if (windx == NULL)
    {
#define READ_INDEX(_i_)   _i_
#define WRITE_INDEX(_i_)  _i_
       DO_INDEXED
#undef READ_INDEX
#undef WRITE_INDEX

    } else
    {
#define READ_INDEX(_i_)   _i_
#define WRITE_INDEX(_i_)  windx[_i_]
       DO_INDEXED
#undef READ_INDEX
#undef WRITE_INDEX
    }

  } else
  {
    if (windx == NULL)
    {
#define READ_INDEX(_i_)   rindx[_i_]
#define WRITE_INDEX(_i_)  _i_
       DO_INDEXED
#undef READ_INDEX
#undef WRITE_INDEX

    } else
    {
#define READ_INDEX(_i_)   rindx[_i_]
#define WRITE_INDEX(_i_)  windx[_i_]
       DO_INDEXED
#undef READ_INDEX
#undef WRITE_INDEX
    }
  }

#undef DO_INDEXED

  return 0;
}


slint_t elements_pack(elements_t *s, packed_elements_t *d) /* sl_proto, sl_func elements_pack */
{
  return elements_pack_indexed(s, d, NULL, NULL);
}


slint_t elements_pack_at(elements_t *s, slint_t sat, packed_elements_t *d, slint_t dat) /* sl_proto, sl_func elements_pack_at */
{
  slint_t i;

  if (s->size > d->size) return -1;

  for (i = 0; i < s->size; ++i)
  {
    key_copy_at(s->keys, sat + i, &d->elements[dat + i].key, 0);
    data_copy_at(s, sat +i, &d->elements[dat + i], 0);
  }

  return 0;
}


slint_t elements_unpack_indexed(packed_elements_t *s, elements_t *d, slindex_t *rindx, slindex_t *windx) /* sl_proto, sl_func elements_unpack_indexed */
{
  slint_t i;

  if (s->size > d->size) return -1;

#define DO_INDEXED \
  for (i = 0; i < s->size; ++i) \
  { \
    key_copy_at(&s->elements[READ_INDEX(i)].key, 0, d->keys, WRITE_INDEX(i)); \
    data_copy_at(&s->elements[READ_INDEX(i)], 0, d, WRITE_INDEX(i)); \
  }

  if (rindx == NULL)
  {
    if (windx == NULL)
    {
#define READ_INDEX(_i_)   _i_
#define WRITE_INDEX(_i_)  _i_
       DO_INDEXED
#undef READ_INDEX
#undef WRITE_INDEX

    } else
    {
#define READ_INDEX(_i_)   _i_
#define WRITE_INDEX(_i_)  windx[_i_]
       DO_INDEXED
#undef READ_INDEX
#undef WRITE_INDEX
    }

  } else
  {
    if (windx == NULL)
    {
#define READ_INDEX(_i_)   rindx[_i_]
#define WRITE_INDEX(_i_)  _i_
       DO_INDEXED
#undef READ_INDEX
#undef WRITE_INDEX

    } else
    {
#define READ_INDEX(_i_)   rindx[_i_]
#define WRITE_INDEX(_i_)  windx[_i_]
       DO_INDEXED
#undef READ_INDEX
#undef WRITE_INDEX
    }
  }

#undef DO_INDEXED

  return 0;
}


slint_t elements_unpack(packed_elements_t *s, elements_t *d) /* sl_proto, sl_func elements_unpack */
{
  return elements_unpack_indexed(s, d, NULL, NULL);
}


slint_t elements_unpack_at(packed_elements_t *s, slint_t sat, elements_t *d, slint_t dat) /* sl_proto, sl_func elements_unpack_at */
{
  slint_t i;

  if (s->size > d->size) return -1;

  for (i = 0; i < s->size; ++i)
  {
    key_copy_at(&s->elements[sat + i].key, 0, d->keys, dat + i);
    data_copy_at(&s->elements[sat + i], 0, d, dat + i);
  }
  
  return 0;
}


slint_t elements_unpack_keys(packed_elements_t *s, slkey_t *k) /* sl_proto, sl_func elements_unpack_keys */
{
  slint_t i;

  for (i = 0; i < s->size; ++i) key_copy_at(&s->elements[i].key, 0, k, i);
  
  return 0;
}
