/*
 *  Copyright (C) 2010, 2011, 2012 Michael Hofmann, Chemnitz University of Technology
 *  
 *  This file is part of the Sorting Library.
 *  
 *  The Sorting Library is free software: you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  The Sorting Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  

 *  
 *  SL - Sorting Library, michael <dot> hofmann <at> informatik <dot> tu-chemnitz <dot> de
 */


#include "sl_common.h"


/*
  ..._lt versions return:
   - max number of elements less than k
   - index i with s->keys[i-1] < k <= s->keys[i]
  ..._le versions return:
   - max number of elements less than or equal k
   - index i with s->keys[i-1] <= k < s->keys[i]
*/


/* sequential search routines, max comparisons: O(n) */

slint sl_search_sequential_lt(elements_t *s, slpkey_t k) /* sl_proto, sl_func sl_search_sequential_lt */
{
  slkey_t *_s, *_e;

  if (s->size <= 0) return 0;

  key_assign(s->keys, _s);
  key_assign_at(s->keys, s->size, _e);

  while (_s != _e)
  if (key_pure_cmp_lt(key_purify(*_s), k)) key_inc(_s); else break;

  return s->size - (_e - _s);
}


slint sl_search_sequential_le(elements_t *s, slpkey_t k) /* sl_proto, sl_func sl_search_sequential_le */
{
  slkey_t *_s, *_e;

  if (s->size <= 0) return 0;

  key_assign(s->keys, _s);
  key_assign_at(s->keys, s->size, _e);

  while (_s != _e)
  if (key_pure_cmp_le(key_purify(*_s), k)) key_inc(_s); else break;

  return s->size - (_e - _s);
}


slint sl_search_sequential_gt(elements_t *s, slpkey_t k) /* sl_proto, sl_func sl_search_sequential_gt */
{
  slkey_t *_s, *_e;

  if (s->size <= 0) return 0;

  key_assign_at(s->keys, s->size - 1, _s);
  key_assign_at(s->keys, -1, _e);

  while (_s != _e)
  if (key_pure_cmp_gt(key_purify(*_s), k)) key_dec(_s); else break;

  return s->size - (_s - _e);
}


slint sl_search_sequential_ge(elements_t *s, slpkey_t k) /* sl_proto, sl_func sl_search_sequential_ge */
{
  slkey_t *_s, *_e;

  if (s->size <= 0) return 0;

  key_assign_at(s->keys, s->size - 1, _s);
  key_assign_at(s->keys, -1, _e);

  while (_s != _e)
  if (key_pure_cmp_ge(key_purify(*_s), k)) key_dec(_s); else break;

  return s->size - (_s - _e);
}


slint sl_search_p_sequential_lt(elements_t *s, slpkey_t *k) /* sl_proto, sl_func sl_search_p_sequential_lt */
{
  return sl_search_sequential_lt(s, *k);
}


slint sl_search_p_sequential_le(elements_t *s, slpkey_t *k) /* sl_proto, sl_func sl_search_p_sequential_le */
{
  return sl_search_sequential_le(s, *k);
}


slint sl_search_p_sequential_gt(elements_t *s, slpkey_t *k) /* sl_proto, sl_func sl_search_p_sequential_gt */
{
  return sl_search_sequential_gt(s, *k);
}


slint sl_search_p_sequential_ge(elements_t *s, slpkey_t *k) /* sl_proto, sl_func sl_search_p_sequential_ge */
{
  return sl_search_sequential_ge(s, *k);
}


/* binary search routines, max comparisons: O(\log_2 n) */

slint sl_search_binary_lt(elements_t *s, slpkey_t k) /* sl_proto, sl_func sl_search_binary_lt */
{
  slint le, ri, mi;

  le = 0;
  ri = s->size - 1;

  while (le <= ri)
  {
    mi = (le + ri) / 2;
    if (key_pure_cmp_le(k, key_purify(s->keys[mi]))) ri = mi - 1;
    else le = mi + 1;
  }

  return le;
}


slint sl_search_binary_le(elements_t *s, slpkey_t k) /* sl_proto, sl_func sl_search_binary_le */
{
  slint le, ri, mi;

  le = 0;
  ri = s->size - 1;

  while (le <= ri)
  {
    mi = (le + ri) / 2;
    if (key_pure_cmp_lt(k, key_purify(s->keys[mi]))) ri = mi - 1;
    else le = mi + 1;
  }

  return le;
}


slint sl_search_binary_gt(elements_t *s, slpkey_t k) /* sl_proto, sl_func sl_search_binary_gt */
{
  return s->size - sl_search_binary_le(s, k);
}


slint sl_search_binary_ge(elements_t *s, slpkey_t k) /* sl_proto, sl_func sl_search_binary_ge */
{
  return s->size - sl_search_binary_lt(s, k);
}


slint sl_search_p_binary_lt(elements_t *s, slpkey_t *k) /* sl_proto, sl_func sl_search_p_binary_lt */
{
  return sl_search_binary_lt(s, *k);
}


slint sl_search_p_binary_le(elements_t *s, slpkey_t *k) /* sl_proto, sl_func sl_search_p_binary_le */
{
  return sl_search_binary_le(s, *k);
}


slint sl_search_p_binary_gt(elements_t *s, slpkey_t *k) /* sl_proto, sl_func sl_search_p_binary_gt */
{
  return sl_search_binary_gt(s, *k);
}


slint sl_search_p_binary_ge(elements_t *s, slpkey_t *k) /* sl_proto, sl_func sl_search_p_binary_ge */
{
  return sl_search_binary_ge(s, *k);
}


slint_t sl_search_binary_lt_bmask(elements_t *s, slpkey_t k, slpkey_t bmask) /* sl_proto, sl_func sl_search_binary_lt_bmask */
{
  slint le, ri, mi;

  le = 0;
  ri = s->size - 1;

  while (le <= ri)
  {
    mi = (le + ri) / 2;
    if (key_pure_cmp_le(k, key_purify(s->keys[mi]) & bmask)) ri = mi - 1;
    else le = mi + 1;
  }

  return le;
}


slint_t sl_search_binary_le_bmask(elements_t *s, slpkey_t k, slpkey_t bmask) /* sl_proto, sl_func sl_search_binary_le_bmask */
{
  slint le, ri, mi;

  le = 0;
  ri = s->size - 1;

  while (le <= ri)
  {
    mi = (le + ri) / 2;
    if (key_pure_cmp_lt(k, key_purify(s->keys[mi]) & bmask)) ri = mi - 1;
    else le = mi + 1;
  }

  return le;
}


slint_t sl_search_binary_sign_switch(elements_t *s) /* sl_proto, sl_func sl_search_binary_sign_switch */
{
  slint le, ri, mi;

  if (s->size <= 0) return 0;
  
  if (key_pure_cmp_ge(key_purify(s->keys[0]), 0) && key_pure_cmp_ge(key_purify(s->keys[s->size - 1]), 0)) return s->size;

  if (key_pure_cmp_lt(key_purify(s->keys[0]), 0) && key_pure_cmp_lt(key_purify(s->keys[s->size - 1]), 0)) return 0;
  
  le = 0;
  ri = s->size - 1;

  while (le <= ri)
  {
    mi = (le + ri) / 2;
    if (key_pure_cmp_ge(key_purify(s->keys[mi]), 0)) le = mi + 1;
    else ri = mi - 1;
  }
  
  return le;
}


/* hybrid search routines combining sequential steps of size t with a following binary search */

slint sl_search_hybrid_lt(elements_t *s, slpkey_t k, slint t) /* sl_proto, sl_func sl_search_hybrid_lt */
{
  slint n;
  elements_t x;
  slkey_t *_s, *_e;

  if (s->size <= 0) return 0;

  key_assign_at(s->keys, t - 1, _s);
  key_assign_at(s->keys, s->size, _e);

  while (_s < _e)
  if (key_pure_cmp_lt(key_purify(*_s), k)) key_add(_s, t); else break;

  n = (_s - s->keys) - (t - 1);

  elem_assign_at(s, n, &x);
  x.size = (z_min(_s, _e) - s->keys) - n;

  return n + sl_search_binary_lt(&x, k);
}


slint sl_search_hybrid_le(elements_t *s, slpkey_t k, slint t) /* sl_proto, sl_func sl_search_hybrid_le */
{
  slint n;
  elements_t x;
  slkey_t *_s, *_e;

  if (s->size <= 0) return 0;

  key_assign_at(s->keys, t - 1, _s);
  key_assign_at(s->keys, s->size, _e);

  while (_s < _e)
  if (key_pure_cmp_le(key_purify(*_s), k)) key_add(_s, t); else break;

  n = (_s - s->keys) - (t - 1);

  elem_assign_at(s, n, &x);
  x.size = (z_min(_s, _e) - s->keys) - n;

  return n + sl_search_binary_le(&x, k);
}


slint sl_search_hybrid_gt(elements_t *s, slpkey_t k, slint t) /* sl_proto, sl_func sl_search_hybrid_gt */
{
  slint n;
  elements_t x;
  slkey_t *_s, *_e;

  if (s->size <= 0) return 0;

  key_assign_at(s->keys, s->size - t, _s);
  key_assign_at(s->keys, -1, _e);

  while (_s > _e)
  if (key_pure_cmp_gt(key_purify(*_s), k)) key_sub(_s, t); else break;

  n = (z_max(_e, _s) - s->keys) + 1;

  elem_assign_at(s, n, &x);
  x.size = ((_s - s->keys) + t) - n;

  return s->size - (n + sl_search_binary_le(&x, k));
}


slint sl_search_hybrid_ge(elements_t *s, slpkey_t k, slint t) /* sl_proto, sl_func sl_search_hybrid_ge */
{
  slint n;
  elements_t x;
  slkey_t *_s, *_e;

  if (s->size <= 0) return 0;

  key_assign_at(s->keys, s->size - t, _s);
  key_assign_at(s->keys, -1, _e);

  while (_s > _e)
  if (key_pure_cmp_ge(key_purify(*_s), k)) key_sub(_s, t); else break;

  n = (z_max(_e, _s) - s->keys) + 1;

  elem_assign_at(s, n, &x);
  x.size = ((_s - s->keys) + t) - n;

  return s->size - (n + sl_search_binary_lt(&x, k));
}


slint sl_search_p_hybrid_lt(elements_t *s, slpkey_t *k, slint t) /* sl_proto, sl_func sl_search_p_hybrid_lt */
{
  return sl_search_hybrid_lt(s, *k, t);
}


slint sl_search_p_hybrid_le(elements_t *s, slpkey_t *k, slint t) /* sl_proto, sl_func sl_search_p_hybrid_le */
{
  return sl_search_hybrid_le(s, *k, t);
}


slint sl_search_p_hybrid_gt(elements_t *s, slpkey_t *k, slint t) /* sl_proto, sl_func sl_search_p_hybrid_gt */
{
  return sl_search_hybrid_gt(s, *k, t);
}


slint sl_search_p_hybrid_ge(elements_t *s, slpkey_t *k, slint t) /* sl_proto, sl_func sl_search_p_hybrid_ge */
{
  return sl_search_hybrid_ge(s, *k, t);
}
