/*
 *  Copyright (C) 2010, 2011, 2012 Michael Hofmann, Chemnitz University of Technology
 *  
 *  This file is part of the Sorting Library.
 *  
 *  The Sorting Library is free software: you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  The Sorting Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  

 *  
 *  SL - Sorting Library, michael <dot> hofmann <at> informatik <dot> tu-chemnitz <dot> de
 */


/* sl_macro SRI_INSERTSORT */
#define SRI_INSERTSORT


#include "sl_common.h"

/* sl_context CONTEXT_BEGIN sri */
const slint_t default_sri_threshold = sort_radix_iter_threshold;  /* sl_global sl_context sl_var default_sri_threshold */
/* sl_context CONTEXT_END sri */


#ifdef key_integer


#define max_nclasses (z_powof2_typed(sort_radix_width_max, slkey_pure_t))


/* insert sort */
static void rs_iter_insertsort(elements_t *s, elements_t *sx, slint_t presorted, slint_t rhigh, slint_t rlow)
{
  slint_t i, j;

  slkey_pure_t class_mask = 0, bmask = 0;


  if (presorted && (rhigh < key_radix_high)) class_mask = ~(z_powof2_typed(rhigh + 1, slkey_pure_t) - 1);

  if (rhigh - rlow + 1 <= key_radix_high) bmask = z_powof2_typed(rhigh - rlow + 1, slkey_pure_t);

  bmask = class_mask | (bmask - 1) << rlow;

  for (i = 1; i < s->size; i++)
  {
    if (key_pure_cmp_lt(key_purify(s->keys[i]) & bmask, key_purify(s->keys[i - 1]) & bmask))
    {
      j = i - 1;
      elem_copy_at(s, i, sx, 0);

      do
      {
        elem_copy_at(s, j, s, j + 1);
        if (--j < 0) break;

      } while (key_pure_cmp_lt(key_purify(*sx->keys) & bmask, key_purify(s->keys[j]) & bmask));

      elem_copy_at(sx, 0, s, j + 1);
    }
  }
}


static void rs_iter(elements_t *s, elements_t *sx, slint_t presorted, slint_t rhigh, slint_t rlow, slint_t rwidth)
{
  slkey_pure_t class_mask, bit_mask, current_class, nclasses;

  slint_t i, j, k, current_width, c[max_nclasses];
  elements_t xi, xj, end, parts[max_nclasses];
  
  slint_t rhigh_old = rhigh;


  if (presorted && (rhigh < key_radix_high)) class_mask = ~(z_powof2_typed(rhigh + 1, slkey_pure_t) - 1);
  else class_mask = 0;

  elem_assign_at(s, s->size, &end);

  while (rhigh >= rlow)
  {
    current_width = z_min(rwidth, rhigh - rlow + 1);
    rhigh -= current_width - 1;

    nclasses = z_powof2_typed(current_width, slkey_pure_t);
    bit_mask = nclasses - 1;

    elem_assign(s, &xi);

    while (xi.keys < end.keys)
    {
      elem_assign(&xi, &xj);

      current_class = class_mask & key_purify(*xi.keys);

      /* zero all counters */
      for (i = 0; i < nclasses; ++i) c[i] = 0;

      /* counting the number of sub-class elements in the current class */
      do
      {
        i = key_radix_key2class(key_purify(*xi.keys), rhigh, bit_mask);

        ++c[i];

        elem_inc(&xi);

        /* stop if we are at the end */
        if (xi.keys >= end.keys) break;

      } while (current_class == (class_mask & key_purify(*xi.keys)));

      /* if the current class has enough elements, perform split */
      if (xi.keys - xj.keys > 
#ifdef SRI_INSERTSORT
        SL_DEFCON(sri.threshold)
#else
        1
#endif
        )
      {
        /* compute the target of every sub-class */
        elem_assign(&xj, &parts[0]);
        for (i = 1; i < nclasses; i++) elem_assign_at(&parts[i - 1], c[i - 1], &parts[i]);;

        for (i = 0; i < nclasses; i++)
        {
          elem_add(&xj, c[i]);
          elem_assign(&parts[i], &xi);

          while (xi.keys < xj.keys)
          {
            j = key_radix_key2class(key_purify(*xi.keys), rhigh, bit_mask);

            while (j != i)
            {
              k = key_radix_key2class(key_purify(*parts[j].keys), rhigh, bit_mask);

              if (k != j) elem_xchange(&xi, &parts[j], sx);

              elem_inc(&parts[j]);

              j = k;
            }

            elem_inc(&xi);
          }
        }
      }
#ifdef SRI_INSERTSORT
        else
      {
        xj.size = xi.keys - xj.keys;

        rs_iter_insertsort(&xj, sx, presorted, rhigh_old, rlow);
      }
#endif
    }

    class_mask |= (bit_mask << rhigh);

    --rhigh;
  }
}


slint_t sort_radix_iter(elements_t *s, elements_t *sx, slint_t presorted, slint_t rhigh, slint_t rlow, slint_t rwidth) /* sl_proto, sl_func sort_radix_iter */
{
  elements_t _sx;

  if (s == NULL) return -1;

  if (s->size < 2) return 0;

  rti_tstart(rti_tid_sort_radix_iter);

  if (sx == NULL || sx->size < 1)
  {
    sx = &_sx;
    elements_alloc(sx, 1, SLCM_ALL);
  }

  if (rhigh < 0) rhigh = key_radix_high;
  if (rlow < 0) rlow = key_radix_low;
  if (rwidth <= 0) rwidth = sort_radix_iter_width_default;

  rs_iter(s, sx, presorted, rhigh, rlow, z_min(rwidth, sort_radix_width_max));

  if (sx == &_sx) elements_free(sx);

  rti_tstop(rti_tid_sort_radix_iter);

  return 0;
}


#undef max_nclasses


#endif /* key_integer */
