/*
 *  Copyright (C) 2010, 2011, 2012 Michael Hofmann, Chemnitz University of Technology
 *  
 *  This file is part of the Sorting Library.
 *  
 *  The Sorting Library is free software: you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  The Sorting Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  

 *  
 *  SL - Sorting Library, michael <dot> hofmann <at> informatik <dot> tu-chemnitz <dot> de
 */


#include "dash_exec_sl.h"


#ifndef DS_TRACE_IF
# define DS_TRACE_IF  (z_mpi_rank == -1)
#endif


dsint_t ds_exec_sl_create(ds_exec_t *exec) /* ds_func ds_exec_sl_create */
{
  ds_exec_sl_t *exec_sl;


  exec->naddrs = 0;

  exec->cxt = exec_sl = z_alloc(1, sizeof(ds_exec_sl_t));

  exec->pre_run = ds_exec_sl_pre_run;
  exec->post_run = ds_exec_sl_post_run;
  exec->make = ds_exec_sl_make;
  exec->move = ds_exec_sl_move;

  return 0;
}


dsint_t ds_exec_sl_destroy(ds_exec_t *exec) /* ds_func ds_exec_sl_destroy */
{
  z_free(exec->cxt);

  return 0;
}


dsint_t ds_exec_sl_pre_run(ds_exec_t *exec) /* ds_func ds_exec_sl_pre_run */
{
  DEFINE_EXEC_SL(exec, exec_sl);


  exec_sl->nmax = exec->ds->sends.nmax + exec->ds->recvs.nmax;
  exec_sl->n = 0;
  exec_sl->reqs = z_alloc(exec_sl->nmax, sizeof(MPI_Request));
  exec_sl->stats = z_alloc(exec_sl->nmax, sizeof(MPI_Status));

  return 0;
}


dsint_t ds_exec_sl_post_run(ds_exec_t *exec) /* ds_func ds_exec_sl_post_run */
{
  DEFINE_EXEC_SL(exec, exec_sl);


  z_free(exec_sl->reqs); exec_sl->reqs = NULL;
  z_free(exec_sl->stats); exec_sl->stats = NULL;

  return 0;
}


dsint_t ds_exec_sl_make(ds_exec_t *exec) /* ds_func ds_exec_sl_make */
{
  DEFINE_EXEC_SL(exec, exec_sl);
  dsint_t i, req_id;
  dsint_t src_buf_id, dst_buf_id;
  elements_t *src_addr, *dst_addr;

  dsint_t first = 1;
  int rcount;


#define xelem_call \
  req_id = 0; \
  for (i = 0; i < exec->ds->recvs.n; ++i) \
  { \
    dst_buf_id = exec->ds->recvs.buf_ids[i]; \
    dst_addr = exec->addrs[exec->ds->sched->bufs[dst_buf_id].addr_id]; \
    MPI_Irecv(xelem_buf_at(dst_addr, exec->ds->recvs.displs[i]), exec->ds->recvs.counts[i], xelem_mpi_datatype, exec->ds->recvs.srcs[i], DS_EXEC_SL_DATA_TAG, exec->ds->comm, &exec_sl->reqs[req_id]); \
    ++req_id; \
  } \
  for (i = 0; i < exec->ds->sends.n; ++i) \
  { \
    src_buf_id = exec->ds->sends.buf_ids[i]; \
    src_addr = exec->addrs[exec->ds->sched->bufs[src_buf_id].addr_id]; \
    MPI_Isend(xelem_buf_at(src_addr, exec->ds->sends.displs[i]), exec->ds->sends.counts[i], xelem_mpi_datatype, exec->ds->sends.dsts[i], DS_EXEC_SL_DATA_TAG, exec->ds->comm, &exec_sl->reqs[req_id]); \
    ++req_id; \
  } \
  for (i = 0; i < exec->ds->locals.n; ++i) \
  { \
    src_buf_id = exec->ds->locals.src_buf_ids[i]; \
    dst_buf_id = exec->ds->locals.dst_buf_ids[i]; \
    src_addr = exec->addrs[exec->ds->sched->bufs[src_buf_id].addr_id]; \
    dst_addr = exec->addrs[exec->ds->sched->bufs[dst_buf_id].addr_id]; \
    if (src_buf_id == dst_buf_id && z_max(exec->ds->locals.src_displs[i], exec->ds->locals.src_displs[i]) < z_min(exec->ds->locals.src_displs[i] + exec->ds->locals.src_counts[i], exec->ds->locals.src_displs[i] + exec->ds->locals.src_counts[i])) \
    { \
      xelem_nmove_at(src_addr, exec->ds->locals.src_displs[i], dst_addr, exec->ds->locals.dst_displs[i], exec->ds->locals.src_counts[i]); \
    } else \
    { \
      xelem_ncopy_at(src_addr, exec->ds->locals.src_displs[i], dst_addr, exec->ds->locals.dst_displs[i], exec->ds->locals.src_counts[i]); \
    } \
  } \
  MPI_Waitall(req_id, exec_sl->reqs, exec_sl->stats); \
  if (first) \
  { \
    for (i = 0; i < exec->ds->recvs.n; ++i) \
    { \
      MPI_Get_count(&exec_sl->stats[i], xelem_mpi_datatype, &rcount); \
      exec->ds->recvs.counts[i] = rcount; \
    } \
    first = 0; \
  }
#include "sl_xelem_call.h"

  return 0;
}


dsint_t ds_exec_sl_add_address(ds_exec_t *exec, elements_t *s) /* ds_func ds_exec_sl_add_address */
{
  return ds_exec_add_address(exec, s);
}


void ds_exec_sl_move(ds_exec_t *exec, dsint_t exec_id, dsint_t src_buf_id, dsint_t src_displs, dsint_t dst_buf_id, dsint_t dst_displs, dsint_t count) /* ds_func ds_exec_sl_move */
{
  elements_t *src_addr, *dst_addr;


  src_addr = exec->addrs[exec->ds->sched->bufs[src_buf_id].addr_id];
  dst_addr = exec->addrs[exec->ds->sched->bufs[dst_buf_id].addr_id];

  if (src_buf_id != dst_buf_id || src_displs + count < dst_displs || dst_displs + count < src_displs)
  {
#define xelem_call \
    xelem_ncopy_at(src_addr, src_displs, dst_addr, dst_displs, count);
#include "sl_xelem_call.h"
  } else
  {
#define xelem_call \
    xelem_nmove_at(src_addr, src_displs, dst_addr, dst_displs, count);
#include "sl_xelem_call.h"
  }
}


#undef DS_TRACE_IF
