/*
 *  Copyright (C) 2010, 2011, 2012 Michael Hofmann, Chemnitz University of Technology
 *  
 *  This file is part of the Sorting Library.
 *  
 *  The Sorting Library is free software: you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  The Sorting Library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  

 *  
 *  SL - Sorting Library, michael <dot> hofmann <at> informatik <dot> tu-chemnitz <dot> de
 */


#include "sl_common.h"


#define N_POWOF2  18


void tune_sort_radix_width(tune_data_t tune_data) /* sl_func tune_sort_radix_width */
{
  const int n = z_powof2(N_POWOF2);
  const int m = sort_radix_width_max;

  int i;
  double trec_db, trec_ip, titer;

  int best_rec_ip, best_rec_db, best_iter;
  double best_trec_ip, best_trec_db, best_titer;

  elements_t s, s1, sx;


  tune_auto_begin("radix-sort default bit widths");

  elements_alloc(&s, n, SLCM_ALL);
  elements_alloc(&s1, n, SLCM_ALL);
  elements_alloc(&sx, n, SLCM_ALL);

  elements_keys_init_seed(2501);
  elements_keys_init(&s, SL_EKIT_RAND, NULL);

  best_trec_ip = best_trec_db = best_titer = 1e99;

  for (i = 1; i <= m; i++)
  {
    elem_ncopy(&s, &s1, s.size);

    trec_ip = z_time_get_s();
    sort_radix_ip(&s1, &sx, -1, -1, i);
    trec_ip = z_time_get_s() - trec_ip;

    if ((best_trec_ip - trec_ip) / trec_ip > -0.005)
    {
      best_rec_ip = i;
      best_trec_ip = trec_ip;
    }

    elem_ncopy(&s, &s1, s.size);

    trec_db = z_time_get_s();
    sort_radix_db(&s1, &sx, -1, -1, i);
    trec_db = z_time_get_s() - trec_db;

    if ((best_trec_db - trec_db) / trec_db > -0.005)
    {
      best_rec_db = i;
      best_trec_db = trec_db;
    }

    elem_ncopy(&s, &s1, s.size);

    titer = z_time_get_s();
    sort_radix_iter(&s1, &sx, 0, -1, -1, i);
    titer = z_time_get_s() - titer;

    if ((best_titer - titer) / titer > -0.005)
    {
      best_iter = i;
      best_titer = titer;
    }

    fprintf(TUNE_PROGRESS_STREAM, "%3d  %f  %f  %f\n", i, trec_ip, trec_db, titer);
  }

  elements_free(&s);
  elements_free(&s1);
  elements_free(&sx);

  tune_auto_end("radix-sort default bit widths");

  tune_auto_output_result_slint("sort_radix_width_max", sort_radix_width_max);
  tune_auto_output_result_slint("sort_radix_ip_width_default", best_rec_ip);
  tune_auto_output_result_slint("sort_radix_db_width_default", best_rec_db);
  tune_auto_output_result_slint("sort_radix_iter_width_default", best_iter);
}


void tune_sort_radix_threshold(tune_data_t tune_data) /* sl_func tune_sort_radix_threshold */
{
  const slint_t n = z_powof2(N_POWOF2);
  const slint_t m = 1024;

  double x[m];
  double trec_ip[m], trec_ip2[m], trec_db[m], trec_db2[m], titer[m], titer2[m];

  slint_t i, old_rec[3], old_iter;
  slint_t rec_ip, rec_db, iter;

  elements_t s, s1, sx;

  
  tune_auto_begin("radix-sort thresholds");

  elements_alloc(&s, n, SLCM_ALL);
  elements_alloc(&s1, n, SLCM_ALL);
  elements_alloc(&sx, n, SLCM_ALL);

  elements_keys_init_seed(2501);
  elements_keys_init(&s, SL_EKIT_RAND, NULL);

  old_rec[0] = SL_DEFCON(sr.ip_threshold);
  old_rec[1] = SL_DEFCON(sr.db_threshold);
  old_rec[2] = SL_DEFCON(sr.ma_threshold);

  old_iter = SL_DEFCON(sri.threshold);

  for (i = 1; i <= m; i++)
  {
    x[i] = i;
  
    SL_DEFCON(sr.ip_threshold) = i;

    elem_ncopy(&s, &s1, i); elem_set_size(&s1, i);

    trec_ip[i] = z_time_get_s();
    sort_radix_ip(&s1, &sx, -1, -1, -1);
    trec_ip[i] = z_time_get_s() - trec_ip[i];

    SL_DEFCON(sr.ip_threshold) = i - 1;

    elem_ncopy(&s, &s1, i); elem_set_size(&s1, i);

    trec_ip2[i] = z_time_get_s();
    sort_radix_ip(&s1, &sx, -1, -1, -1);
    trec_ip2[i] = z_time_get_s() - trec_ip2[i];

    SL_DEFCON(sr.db_threshold) = i;

    elem_ncopy(&s, &s1, i); elem_set_size(&s1, i);

    trec_db[i] = z_time_get_s();
    sort_radix_db(&s1, &sx, -1, -1, -1);
    trec_db[i] = z_time_get_s() - trec_db[i];

    SL_DEFCON(sr.db_threshold) = i - 1;

    elem_ncopy(&s, &s1, i); elem_set_size(&s1, i);

    trec_db2[i] = z_time_get_s();
    sort_radix_db(&s1, &sx, -1, -1, -1);
    trec_db2[i] = z_time_get_s() - trec_db2[i];

    SL_DEFCON(sri.threshold) = i;

    elem_ncopy(&s, &s1, i); elem_set_size(&s1, i);

    titer[i] = z_time_get_s();
    sort_radix_iter(&s1, &sx, 0, -1, -1, -1);
    titer[i] = z_time_get_s() - titer[i];

    SL_DEFCON(sri.threshold) = i - 1;

    elem_ncopy(&s, &s1, i); elem_set_size(&s1, i);

    titer2[i] = z_time_get_s();
    sort_radix_iter(&s1, &sx, 0, -1, -1, -1);
    titer2[i] = z_time_get_s() - titer2[i];

    fprintf(TUNE_PROGRESS_STREAM, "%3" slint_fmt "  %f  %f  %f  %f  %f  %f\n", i, trec_ip[i], trec_ip2[i], trec_db[i], trec_db2[i], titer[i], titer2[i]);
  }

  rec_ip = z_round(intersect_poly2(m, x, trec_ip, trec_ip2, 1, m));
  rec_db = z_round(intersect_poly2(m, x, trec_db, trec_db2, 1, m));
  iter = z_round(intersect_poly2(m, x, titer, titer2, 1, m));

  SL_DEFCON(sr.ip_threshold) = old_rec[0];
  SL_DEFCON(sr.db_threshold) = old_rec[1];
  SL_DEFCON(sr.ma_threshold) = old_rec[2];

  SL_DEFCON(sri.threshold) = old_iter;

  elements_free(&s);
  elements_free(&s1);
  elements_free(&sx);

  tune_auto_end("radix-sort thresholds");

  tune_auto_output_result_slint("sort_radix_ip_threshold", rec_ip);
  tune_auto_output_result_slint("sort_radix_db_threshold", rec_db);
  tune_auto_output_result_slint("sort_radix_iter_threshold", iter);
}
